package de.fraunhofer.sit.c2x.pki;

import java.io.IOException;
import java.security.Security;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.bouncycastle.crypto.AsymmetricCipherKeyPair;
import org.bouncycastle.crypto.params.ECPrivateKeyParameters;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;

import de.fraunhofer.sit.c2x.pki.ca.certificates.datacontainers.GeographicRegionDataContainer;
import de.fraunhofer.sit.c2x.pki.ca.certificates.datacontainers.PsidSspPriorityDataContainer;
import de.fraunhofer.sit.c2x.pki.ca.core.exceptions.HandlerException;
import de.fraunhofer.sit.c2x.pki.ca.provider.ProviderException;

public class CertChainGenerator {
	private static CertChainGenerator instance;
	private static final DateFormat df = new SimpleDateFormat("dd.MM.yyyy");
	private static final String today = df.format(new Date());

	@Inject
	CertificateGenerator caCertificateRequestHandler;

	public static CertChainGenerator getInstance() {
		if (instance == null)
			instance = new CertChainGenerator();
		return instance;
	}

	public CertChainGenerator() {
		Security.addProvider(new BouncyCastleProvider());
		Injector injector = Guice.createInjector(new CertificateBuilderModule());
		caCertificateRequestHandler = injector.getInstance(CertificateGenerator.class);
	}
	
	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChain(String endDate, int assuranceLevel, long[] aidList,
			byte[][] sspList, byte[] pseudonymSigningKeyX, byte[] pseudonymSigningKeyY) throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList, null, null,
				null, null, null, null, null, null, null, null, null, null, pseudonymSigningKeyX,
				pseudonymSigningKeyY, null, null);
	}

	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChain(String endDate, int assuranceLevel, long[] aidList,
			byte[][] sspList, byte[] pseudonymSigningKeyX, byte[] pseudonymSigningKeyY,
			byte[] pseudonymEncryptionKeyX, byte[] pseudonymEncryptionKeyY) throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList, null, null,
				null, null, null, null, null, null, null, null, null, null, pseudonymSigningKeyX,
				pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY);
	}

	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param circularRegionRestrictionCenterLatitude
	 *            as WGS 84 encoded latitude value
	 * @param circularRegionRestrictionCenterLongitude
	 *            as WGS 84 encoded longitude value
	 * @param circularRegionRestrictionRadius
	 *            in meters
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChainWithCircularRegionRestriction(String endDate,
			int assuranceLevel, long[] aidList, byte[][] sspList,
			double circularRegionRestrictionCenterLatitude, double circularRegionRestrictionCenterLongitude,
			int circularRegionRestrictionRadius, byte[] pseudonymSigningKeyX, byte[] pseudonymSigningKeyY,
			byte[] pseudonymEncryptionKeyX, byte[] pseudonymEncryptionKeyY) throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList,
				circularRegionRestrictionCenterLatitude, circularRegionRestrictionCenterLongitude,
				circularRegionRestrictionRadius, null, null, null, null, null, null, null, null, null,
				pseudonymSigningKeyX, pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY);
	}

	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param rectangularRegionRestrictionNorthWestPointsLatitude
	 *            as WGS 84 encoded latitude value
	 * @param rectangularRegionRestrictionNorthWestPointsLongitude
	 *            as WGS 84 encoded longitude value
	 * @param rectangularRegionRestrictionSouthEastPointsLatitude
	 *            as WGS 84 encoded latitude value
	 * @param rectangularRegionRestrictionSouthEastPointsLongitude
	 *            as WGS 84 encoded longitude value
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChainWithRectangularRegionRestriction(String endDate,
			int assuranceLevel, long[] aidList, byte[][] sspList,
			Double[] rectangularRegionRestrictionNorthWestPointsLatitude,
			Double[] rectangularRegionRestrictionNorthWestPointsLongitude,
			Double[] rectangularRegionRestrictionSouthEastPointsLatitude,
			Double[] rectangularRegionRestrictionSouthEastPointsLongitude, byte[] pseudonymSigningKeyX,
			byte[] pseudonymSigningKeyY, byte[] pseudonymEncryptionKeyX, byte[] pseudonymEncryptionKeyY)
			throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList, null, null,
				null, rectangularRegionRestrictionNorthWestPointsLatitude,
				rectangularRegionRestrictionNorthWestPointsLongitude,
				rectangularRegionRestrictionSouthEastPointsLatitude,
				rectangularRegionRestrictionSouthEastPointsLongitude, null, null, null, null, null,
				pseudonymSigningKeyX, pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY);
	}

	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param polygonalRegionRestrictionPointsLatitude
	 *            as WGS 84 encoded latitude value
	 * @param polygonalRegionRestrictionPointsLongitude
	 *            as WGS 84 encoded longitude value
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChainWithPolygonalRegionRestriction(String endDate,
			int assuranceLevel, long[] aidList, byte[][] sspList,
			Double[] polygonalRegionRestrictionPointsLatitude,
			Double[] polygonalRegionRestrictionPointsLongitude, byte[] pseudonymSigningKeyX,
			byte[] pseudonymSigningKeyY, byte[] pseudonymEncryptionKeyX, byte[] pseudonymEncryptionKeyY)
			throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList, null, null,
				null, null, null, null, null, polygonalRegionRestrictionPointsLatitude,
				polygonalRegionRestrictionPointsLongitude, null, null, null, pseudonymSigningKeyX,
				pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY);
	}

	/**
	 * @param endDate
	 *            as string with the format: dd.mm.yyyy
	 * @param assuranceLevel
	 *            as value: 0,1,2,3,4,5,6,7
	 * @param aidList
	 *            as array of long values giving the Application IDs according
	 *            to ETSI TS 103 097 and ETSI TR 102 965
	 * @param sspList
	 *            as array of byte arrays giving the Service Specific
	 *            Permissions to every AID according to ETSI TS 103 097
	 * @param identifiedRegionRestrictionDictionary
	 * @param identifiedRegionRestrictionIdentifer
	 * @param identifiedRegionRestrictionLocalRegion
	 * @param PseudonymSigningKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymSigningKeyY
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyX
	 *            of NIST p256 ECC public key
	 * @param PseudonymEncryptionKeyY
	 *            of NIST p256 ECC public key
	 * @return array of HEX encoded ETSI TS 103 097 certificates. First
	 *         certificate is the root certificate, second is the Authorization
	 *         Authority certificate and the third is the Authorization Ticket
	 *         certificate
	 * @throws HandlerException
	 */
	public static ArrayList<String> getCertChainWithIdentifiedRegionRestriction(String endDate,
			int assuranceLevel, long[] aidList, byte[][] sspList,
			String identifiedRegionRestrictionDictionary, long identifiedRegionRestrictionIdentifer,
			long identifiedRegionRestrictionLocalRegion, byte[] pseudonymSigningKeyX,
			byte[] pseudonymSigningKeyY, byte[] pseudonymEncryptionKeyX, byte[] pseudonymEncryptionKeyY)
			throws HandlerException {
		return getInstance()._getCertChain(today, endDate, assuranceLevel, aidList, sspList, null, null,
				null, null, null, null, null, null, null, identifiedRegionRestrictionDictionary,
				identifiedRegionRestrictionIdentifer, identifiedRegionRestrictionLocalRegion,
				pseudonymSigningKeyX, pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY);
	}

	private ArrayList<String> _getCertChain(String startDate, String endDate, int assuranceLevel,
			long[] aidList, byte[][] sspList, Double circularRegionRestrictionCenterLatitude,
			Double circularRegionRestrictionCenterLongitude, Integer circularRegionRestrictionRadius,
			Double[] rectangularRegionRestrictionNorthWestPointsLatitude,
			Double[] rectangularRegionRestrictionNorthWestPointsLongitude,
			Double[] rectangularRegionRestrictionSouthEastPointsLatitude,
			Double[] rectangularRegionRestrictionSouthEastPointsLongitude,
			Double[] polygonalRegionRestrictionPointsLatitude,
			Double[] polygonalRegionRestrictionPointsLongitude, String identifiedRegionRestrictionDictionary,
			Long identifiedRegionRestrictionIdentifer, Long identifiedRegionRestrictionLocalRegion,
			byte[] pseudonymSigningKeyX, byte[] pseudonymSigningKeyY, byte[] pseudonymEncryptionKeyX,
			byte[] pseudonymEncryptionKeyY) throws HandlerException {
		ArrayList<String> certChain = new ArrayList<>();
		try {
			// copy permissions (only AIDs for CA certificates) into correct
			// format
			ArrayList<PsidSspPriorityDataContainer> permissionsCA = new ArrayList<>();
			for (int i = 0; i < aidList.length; i++) {
				PsidSspPriorityDataContainer c = new PsidSspPriorityDataContainer();
				c.setPsid(new Long(aidList[i]));
				permissionsCA.add(c);
			}

			ArrayList<GeographicRegionDataContainer> regionRestrictions = new ArrayList<>();
			if (circularRegionRestrictionCenterLatitude != null
					&& circularRegionRestrictionCenterLongitude != null
					&& circularRegionRestrictionRadius != null) {
				// circular region restriction
				GeographicRegionDataContainer r = new GeographicRegionDataContainer(
						circularRegionRestrictionCenterLatitude, circularRegionRestrictionCenterLongitude,
						circularRegionRestrictionRadius);
				regionRestrictions.add(r);
			} else if (rectangularRegionRestrictionNorthWestPointsLatitude != null
					&& rectangularRegionRestrictionNorthWestPointsLongitude != null
					&& rectangularRegionRestrictionSouthEastPointsLatitude != null
					&& rectangularRegionRestrictionSouthEastPointsLongitude != null) {
				// rectangular region restriction
				GeographicRegionDataContainer r = new GeographicRegionDataContainer(
						rectangularRegionRestrictionNorthWestPointsLatitude,
						rectangularRegionRestrictionNorthWestPointsLongitude,
						rectangularRegionRestrictionSouthEastPointsLatitude,
						rectangularRegionRestrictionSouthEastPointsLongitude);
				regionRestrictions.add(r);
			} else if (polygonalRegionRestrictionPointsLatitude != null
					&& polygonalRegionRestrictionPointsLongitude != null) {
				// polygonal region restriction
				GeographicRegionDataContainer r = new GeographicRegionDataContainer(
						polygonalRegionRestrictionPointsLatitude, polygonalRegionRestrictionPointsLongitude);
				regionRestrictions.add(r);
			} else if (identifiedRegionRestrictionDictionary != null
					&& identifiedRegionRestrictionIdentifer != null) {
				// identified region restriction
				GeographicRegionDataContainer r = new GeographicRegionDataContainer(
						identifiedRegionRestrictionDictionary, identifiedRegionRestrictionIdentifer,
						identifiedRegionRestrictionLocalRegion);
				regionRestrictions.add(r);
			}

			// create RCA certificate
			String rcaCert = caCertificateRequestHandler.createRootCert(startDate, endDate, new Integer(
					assuranceLevel).toString(), "ETSI_Plugtest_Root", permissionsCA, regionRestrictions);
			certChain.add(rcaCert);
			AsymmetricCipherKeyPair rcaSigningKeyPair = caCertificateRequestHandler.getSigningKeyPair();

			// create PCA certificate
			String pcaCert = caCertificateRequestHandler.createPCACert(startDate, endDate, permissionsCA,
					(ECPrivateKeyParameters) rcaSigningKeyPair.getPrivate(),
					Hex.decodeHex(rcaCert.toCharArray()), new Integer(assuranceLevel).toString(),
					"ETSI_Plugtest_AA", regionRestrictions);
			certChain.add(pcaCert);
			AsymmetricCipherKeyPair pcaSigningKeyPair = caCertificateRequestHandler.getSigningKeyPair();
			byte[] pcaCertBytes = Hex.decodeHex(pcaCert.toCharArray());

			// copy permissions into correct format for pseudonym certs
			ArrayList<PsidSspPriorityDataContainer> permissionsPC = new ArrayList<>();
			for (int i = 0; i < aidList.length; i++) {
				PsidSspPriorityDataContainer c = new PsidSspPriorityDataContainer();
				c.setPsid(new Long(aidList[i]));
				if (sspList.length > i) {
					c.setServiceSpecificPermissions(sspList[i]);
				}
				permissionsPC.add(c);
			}

			// create pseudonym certificate
			String pcCert = caCertificateRequestHandler.createPseudonymCert(pseudonymSigningKeyX,
					pseudonymSigningKeyY, pseudonymEncryptionKeyX, pseudonymEncryptionKeyY, startDate,
					endDate, permissionsPC, (ECPrivateKeyParameters) pcaSigningKeyPair.getPrivate(),
					pcaCertBytes, new Integer(assuranceLevel).toString(), regionRestrictions);
			certChain.add(pcCert);

		} catch (HandlerException | IOException | ProviderException | DecoderException e) {
			e.printStackTrace();
			throw new HandlerException(e);
		}
		return certChain;
	}

}
